/**
 Terrain Controller-- flies around a terrain.
 
 Orion Sky Lawlor, olawlor@acm.org, 2004/4/6
*/
#include "osl/raster.h" /* for RgbaRaster */
#include "ogl/ext.h"
#include "terrain/controller.h"
#include "terrain/renderer.h"


TerrainController::TerrainController(TerrainSet *t_) 
	:t(t_)
{
	win_size=800;
	Vector2d mapCenter=t->getCenter();
	eyePoint=mapCenter+Vector2d(0.001,0.001); /* HACK: prevent roundoff by not sitting on edge */
	eyeHeight=5.0e2; // start 500m above dem
	eyeHeading=0.5*M_PI; // ... facing north
	eyeDip=0;         // ... looking down
	eyeFOV=90.0*(M_PI/180.0);
	state=stateInvalid;
	drawMode=0;
	errTol=1.0; /* Tolerate 1.0 pixels of error */
	useSaved=false;
	compass=0;
	sprintf(msgBuf,"Started up.");
}

/* Control */
void TerrainController::getZRange(double &z_near,double &z_far) {
	z_near=0.5; z_far=500.0e3+10*eyeHeight;
}

void TerrainController::mouseDown(int button,int x,int y) {
	lastX=x; lastY=y;
	if (button==0) state=stateTrans;
	else /* button != 0 */ state=stateRotate;
}

void clipDip(double &eyeDip) {
	const double dipL=-M_PI*0.5+1.0e-4, dipH=0.5; /* <- don't let people look straight up--they'll get lost in the sky... */
	if (eyeDip<dipL) eyeDip=dipL;
	if (eyeDip>dipH) eyeDip=dipH;
}
void TerrainController::mouseDrag(int buttonMask,int x,int y) {
	double dx=x-lastX, dy=y-lastY;
	double dragSpeed=1.0/win_size;
	switch (state) {
	case stateRotate: {
		eyeHeading+=dragSpeed*dx*eyeFOV;
		eyeDip-=dragSpeed*dy*eyeFOV;
		clipDip(eyeDip);
		} break;
	case stateTrans: {
		double s=3*eyeHeight*dragSpeed*eyeFOV;
		eyePoint-=dy*s*eyeDir2d()+dx*s*eyeRight2d();
		} break;
	default: break;
	};
	oglRepaint(10);
	lastX=x; lastY=y;
}
void TerrainController::mouseUp(int buttonMask,int x,int y) {
	state=stateInvalid;
}


/// Render this camera's field of view into the scene, using bounding lines.
void drawFOV(const osl::Viewpoint &vp)
{
	glBegin(GL_LINES);
	for (int y=0;y<2;y++)
	for (int x=0;x<2;x++) {
		osl::Ray ray=vp.getPixelRay(osl::Vector2d(x*vp.getWidth(),y*vp.getHeight()));
		glVertex3dv((const double *)ray.at(0.1));
		glVertex3dv((const double *)ray.at(100000));
	}
	glEnd();
}

Vector3d to3d(Vector2d v) {return Vector3d(v.x,v.y,0);}

void TerrainController::draw(const osl::Viewpoint &v) {
	TexturedTerrainRenderer tex;
	drawTo(v,&tex);
}
void TerrainController::drawTo(const osl::Viewpoint &vp,TerrainRenderer *renderer)
{
	if (!useSaved) savedView=vp;
	TerrainCamera cam(savedView,errTol,drawMode);
	cam.renderer=renderer;
	t->draw(cam);
	if (drawMode&drawMode_stats) cam.viewStats();

	/* Turn off any programmable shaders-- not needed for compass */
	if (glUseProgramObjectARB) {
		glUseProgramObjectARB(0);
	}
	
	if (useSaved) { /* Draw camera FOV into scene */
		glColor4f(1,0,0,0.7);
		drawFOV(savedView);
	}
	// Draw status line on top of screen
	if (0)
	{
		glColor4f(1,0.9,0.8,1);
		oglPixelMatrixSentry sen;
		oglPrint(osl::Vector3d(-0.95,0.9,0),msgBuf);  
	}
	// Draw compass at bottom of screen
	/* Can't actually load textures until GL is started... */
	if (!compass) compass=makeOglTextureNULL("compass.png");
	if (compass) {
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		oglCheck("Before binding compass state");
		oglPixelMatrixSentry sen;
		compass->bind();
		oglTextureType(oglTexture_linear);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glDepthMask(GL_FALSE); /* disable depth writes */
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glEnable(GL_TEXTURE_2D);
		glColor4f(1,1,1,1);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		//glEnable(GL_ALPHA_TEST);
		oglVector3d cen(0.9,-0.98,0); /* center */
		double s=sin(eyeHeading), c=cos(eyeHeading);
		oglVector3d rt(s,c,0.0), up(-c,s,0.0);
		double xs=0.1; /* size onscreen (screen fraction) */
		double ys=xs*vp.getWidth()/vp.getHeight();
		rt.x*=xs; rt.y*=ys;
		up.x*=xs; up.y*=ys;
		oglTextureQuad(
			cen-up-rt,cen-up+rt,
			cen+up-rt,cen+up+rt);
		oglCheck("After binding texture state");
		glPopAttrib();
	}
}

void TerrainController::flipBit(int drawMode_flip,const char *what) {
	drawMode^=drawMode_flip;
	sprintf(msgBuf,"%s %s",what,(drawMode&drawMode_flip)?"enabled":"disabled");
	oglRepaint(10);
}
void TerrainController::setHt(double h) {
	eyeHeight=h;
	if (eyeHeight<500.0)
		sprintf(msgBuf,"Altitude %.1f m",eyeHeight);
	else
		sprintf(msgBuf,"Altitude %.1f km",1.0e-3*eyeHeight);
	oglRepaint(10);
}
void TerrainController::setErr(double e) {
	errTol=e;
	sprintf(msgBuf,"Error tolerance set to %.2f",errTol); 
	oglRepaint(10);
}
void TerrainController::handleEvent(int uiCode) {
	double keySpeed=0.1*eyeHeight;
	switch(uiCode) {
	case 'l': flipBit(drawMode_frame,"Line draw"); break;
	case 'n': flipBit(drawMode_number,"Number draw"); break;
	case 'b': flipBit(drawMode_backface,"Backface draw"); break;
	case 'c': flipBit(drawMode_noclip,"Unclipped view"); break;
	//case 'C': flipBit(drawMode_crack,"Crack view"); break;
	case 't': flipBit(drawMode_notex,"Untextured view"); break;
	case 'm': flipBit(drawMode_nomip,"Unmipmapped view"); break;
	case 'p': flipBit(drawMode_stats,"Statistics prints"); break;
	case '9': flipBit(drawMode_noobjs,"Object skipping"); break;
	case 'v': useSaved=!useSaved; 
		sprintf(msgBuf,"Viewpoint %s",useSaved?"saved":"free"); oglRepaint(10); break;
	case 'e': setErr(2*errTol); break;
	case 'r': setErr(1.0); break;
	case '0': setErr(0.0); break;
	
	case 'z': flipBit(drawMode_zbuf,"Z-Buffer"); break;
//	case 'l': flipBit(drawMode_vieworder,"View order display"); break;
	
	// Camera control:
	case '+': case '=': 
		demoMode=false; setHt(eyeHeight*0.9); break;
	case '-': case '_':
		demoMode=false; setHt(eyeHeight*1.1); break;
	case 'w': 
		demoMode=false; eyePoint+=keySpeed*eyeDir2d(); break;
	case 's': 
		demoMode=false; eyePoint-=keySpeed*eyeDir2d(); break;
	case 'd': 
		demoMode=false; eyePoint+=keySpeed*eyeRight2d(); break;
	case 'a': 
		demoMode=false; eyePoint-=keySpeed*eyeRight2d(); break;
	case 'i': case 'I':
		eyeFOV*=0.9; break;
	case 'o': case 'O':
		eyeFOV*=1.1; break;
	
	// Automated demo mode:
	case 'C':
		demoMode=!demoMode; demoTarget=Vector3d(eyePoint.x,eyePoint.y,eyeHeight); break;
	
	default:
		oglController::handleEvent(uiCode);
		break;
	};
	oglRepaint(10);
};


// oglController interface:
const osl::Viewpoint &TerrainController::getViewpoint(int width,int height)
{
/* Check for joystick events */
	static double last_time=oglTimer();
	double cur_time=oglTimer();
	double dt=cur_time-last_time;
	if (dt>0.1) dt=0.01; /* avoids popping when you don't press for a while */
	last_time=cur_time;
	double zoomSpeed=2.0; /* Zoom fraction/sec */
	if (oglButton(3,"Move Higher")) setHt((1.0+zoomSpeed*dt)*eyeHeight);
	if (oglButton(4,"Move Lower")) setHt((1.0-zoomSpeed*dt)*eyeHeight);
	
	double moveSpeed=3.0*eyeHeight; /* speed in m/s */
	eyePoint+=moveSpeed*dt*(oglAxis(1,"X Move")*eyeRight2d()-oglAxis(2,"Y Move")*eyeDir2d());
	double lookSpeed=1.0; /* radians/s */
	eyeHeading-=lookSpeed*dt*oglAxis(4,"Horizontal Look");
	eyeDip-=lookSpeed*dt*oglAxis(3,"Vertical Look");
	clipDip(eyeDip);

/* Built a viewpoint */
	win_size=width;
	
	Vector2d relEye=eyePoint;
	osl::Vector3d eye(relEye.x,relEye.y,eyeHeight+
		t->getHeight(eyePoint,eyeHeight));
		// ht->at(ht->clipPixel(ht->pixelFmMap(eyePoint))));
	c_viewpoint=osl::Viewpoint(eye,eye+(double *)lookDir(),osl::Vector3d(0,0,1));
	c_viewpoint.discretize(width,height,eyeFOV*180.0/M_PI);

	return c_viewpoint;
}

void TerrainController::timeTests(void)
{ /* run timing test */
	drawMode=0;
	printf("\n-- No display at all\n");
	timeTest(1024,0,1);
	printf("\n-- Ordinary, textured view\n");
	timeTest(1024,0,0);
if (0) {
	printf("\n-- No objects\n");
	timeTest(1024,drawMode_noobjs,0);
	printf("\n-- No view clip\n");
	timeTest(1024,drawMode_noclip,0);
	printf("\n-- Mega view\n");
	timeTest(2048,0,0);
	printf("\n-- No texture\n");
	timeTest(1024,drawMode_notex,0);
}
	exit(0);
}

/** 
 Run a timing test by flying a circle over this terrain.
*/
void TerrainController::timeTest(int wid,int drawModeAdd,int skipDisplay) {
	Vector2d cen2d=t->getCenter();
	double r=0.3*(cen2d-t->getOrigin()).x;
	int nFrames=20;
	double elapsed=0;
	TerrainRenderer *renderer;
	if (skipDisplay) renderer=new TerrainRenderer;
	else renderer=new TexturedTerrainRenderer;
	drawMode|=drawModeAdd;
	for (int f=0;f<nFrames;f++) {
		if (f==0) drawMode|=drawMode_stats; 
		else drawMode&=~drawMode_stats; 
		Vector2d orient(Polar2d(0.4*r,2*M_PI*f/nFrames));
		Vector2d camLoc(cen2d+orient);
		eyeHeight=100.0; // meters above ground
		double camZ=t->getHeight(camLoc,eyeHeight);
		osl::Viewpoint vp(osl::Vector3d(camLoc.x,camLoc.y,camZ+eyeHeight),
			osl::Vector3d(cen2d.x,cen2d.y,t->getHeight(cen2d,0)),
			osl::Vector3d(0,0,1));
		vp.discretize(wid,wid/4*3,70.0);
		oglStartFrame(vp);
		double startTime=oglTimer();
		drawTo(vp,renderer);
		elapsed+=(oglTimer()-startTime)/nFrames;
		
		oglEndFrame();
	}
	delete renderer;
	printf("%d frames, %.2f ms/frame\n",nFrames,1000.0*elapsed);
}

