/*
Primitive rendering for terrain code.
Tree traversal is in terrain.cpp, not here.

Orion Sky Lawlor, olawlor@acm.org, 2004/10/28
*/
#include "terrain/terrain.h"
using namespace osl;

/************ Basic TerrainRenderer ************/
TerrainRenderer::~TerrainRenderer() {}

void TerrainRenderer::beginTerrain(TerrainCamera &cam_,const TerrainSet *s) {
	cam=&cam_;
}
void TerrainRenderer::endTerrain(void) {
	cam=0;
}

void TerrainRenderer::beginTile(const TerrainAddress &addr) {}
void TerrainRenderer::endTile(const TerrainAddress &addr) {}

void TerrainRenderer::expand(const TerrainLevel *b,bool viewCull,int x,int y)
{
	b->expand(*cam,viewCull,x,y);
}

void TerrainRenderer::drawQuad(const TerrainLevel *b,const TerrainAddress &addr) 
{
	const TerrainQuad &q=b->getQuad(addr.x,addr.y);
	if (q.objectList) drawObjects(b,addr,q.objectList);
}
void TerrainRenderer::drawTriangle(const TerrainLevel *b,const TerrainAddress &addr,int v1,int v2) {}
void TerrainRenderer::drawObjects(const TerrainLevel *b,const TerrainAddress &addr,TerrainObjectListID list) {}

/************ Textured OpenGL TerrainRenderer ************/

/**** Management ***/
void TerrainDrawStart(TerrainCamera &cam) {
	glEnable(GL_BLEND);
	// Premultiplied alpha is standard for (my) images
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	
	if (cam.terrainTex) {
		cam.terrainTex->bind();
		glEnable(GL_TEXTURE_2D);
	}
	glEnable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
}

void TerrainDrawStop(void) {
	glDisable(GL_CULL_FACE);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
}


/// Temporarily stop drawing terrain, prepare to draw objects.
///  Turns off the texture coordinate generation used for terrain.
void TerrainDrawPause(TerrainCamera &cam)
{
	TerrainDrawStop();
}

/// Resume drawing terrain after drawing objects.
///  Turns on the texture coordinate generation used for terrain.
void TerrainDrawResume(TerrainCamera &cam)
{
	TerrainDrawStart(cam);
}

/****************** OpenGL Vertices *****************/

static GLenum lazyState=0;
/*
 "Lazy" OpenGL begin/end pairs--only actually ends when needed.
*/
/* Close any pending OpenGL calls */
static void lazyFlush(void) {
	if (lazyState!=0) {
		glEnd();
		lazyState=0;
	}
}
/* Start a glBegin of this operation */
static void lazyBegin(GLenum what) {
	if (lazyState!=what) {
		lazyFlush();
		glBegin(what);
		lazyState=what;
	}
}


void VertexTerrainRenderer::endTile(const TerrainAddress &addr)
{
	flush();
}
void VertexTerrainRenderer::flush(void) {
	lazyFlush();
}

void VertexTerrainRenderer::drawTriangle(const TerrainLevel *b,const TerrainAddress &addr,int v1,int v2)
{
	typedef struct {
		unsigned char dx,dy;
	} vertex_off_t;
	// Convert vertex number to offset:
	const static vertex_off_t vtx[9]={
	   {0,0}, {1,0}, {2,0}, {2,1}, {2,2}, {1,2}, {0,2}, {0,1}, {1,1}
	};
	int x=addr.x, y=addr.y;
	if (cam->getDrawMode()&drawMode_frame) 
	{
		glBegin(GL_LINE_LOOP);
		b->glVertex(x+1,y+1);
		b->glVertex(x+vtx[v1].dx, y+vtx[v1].dy);
		b->glVertex(x+vtx[v2].dx, y+vtx[v2].dy);
		glEnd();
	}
	else { /* regular fill */
		lazyBegin(GL_TRIANGLES);
		b->glVertex(x+1,y+1);
		b->glVertex(x+vtx[v1].dx, y+vtx[v1].dy);
		b->glVertex(x+vtx[v2].dx, y+vtx[v2].dy);
	}
	
	cam->view(TerrainCamera::viewPolys);
	cam->view(TerrainCamera::viewVertices,3);
}

void VertexTerrainRenderer::drawQuad(const TerrainLevel *b,const TerrainAddress &addr) 
{
	int x=addr.x, y=addr.y;
	const TerrainQuad &q=b->getQuad(x,y);
	bool split=q.flags&1;
	const int d=1; // size of quad
	if (cam->getDrawMode()&drawMode_frame) { /* outline triangle decomposition */
		glBegin(GL_LINES);
		if (split) {
		  b->glVertex(x  ,y  );
		  b->glVertex(x+d,y+d);
		} else {
		  b->glVertex(x+d,y  );
		  b->glVertex(x  ,y+d);
		}
		glEnd();
		glBegin(GL_LINE_LOOP);
		b->glVertex(x  ,y  );
		b->glVertex(x+d,y  );
		b->glVertex(x+d,y+d);
		b->glVertex(x  ,y+d);
		glEnd();
	}
	else /* normal fill */
	{
		lazyBegin(GL_TRIANGLES);
		if (split) { /* aiii!!!  duplicate vertices!!!! */
			b->glVertex(x  ,y  );
			b->glVertex(x+d,y  );
			b->glVertex(x+d,y+d);
			
			b->glVertex(x  ,y  );
			b->glVertex(x+d,y+d);
			b->glVertex(x  ,y+d);
		} else {
			b->glVertex(x  ,y  );
			b->glVertex(x+d,y  );
			b->glVertex(x  ,y+d);
			
			b->glVertex(x  ,y+d);
			b->glVertex(x+d,y  );
			b->glVertex(x+d,y+d);
		}
	}
	if (q.objectList) drawObjects(b,addr,q.objectList);
	cam->view(TerrainCamera::viewQuads);
}

/************************* Textures ********************/
TerrainTexture::TerrainTexture(const oglTexture *tex_,
	const Vector2d &texmin_,const Vector2d &texmax_)
		:tex(tex_), texmin(texmin_), texmax(texmax_) {}

TerrainTexture::TerrainTexture(const oglTexture *tex_,
	const GeoImage &texGeo)
	:tex(tex_), 
	 texmin(texGeo.mapFmPixel(0,0)), 
	 texmax(texGeo.mapFmPixel(texGeo.width,texGeo.height)) {}


/// begin using our texture on this camera.
void TerrainTexture::bind(TerrainCamera &cam) const
{
	tex->bind(); // openGL bind
	
	int drawMode=cam.getDrawMode();
	if ( (drawMode&drawMode_notex) 
		|| (drawMode&drawMode_frame) ) 
	{ /* texture disabled */
		glDisable(GL_TEXTURE_2D);
		glColor4f(1,1,1,0.3);
		if (drawMode&drawMode_frame) glDisable(GL_DEPTH_TEST);
		cam.terrainTex=NULL;
	}
	else /* texture enabled */ {
		oglCheck("before texture setup");
		glColor4f(1,1,1,1);
		glEnable(GL_TEXTURE_2D);
		cam.terrainTex=this;
	}
	
	// Texture sampling mode:
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	if (1 || cam.getDrawMode()&drawMode_nomip)
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	else {
        	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
		// glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_LOD, -50.0); 
		oglTexMaxMip(3.0);
	}
	
	// Texture coordinate generation: locations are in heightmap samples.
	for (int axis=0;axis<2;axis++) {
		// (texGenParams dot location) gives normalized texture coords
		//   along each texture axis.
		// We want 
		//    texGenParams dot texmin == 0
		//    texGenParams dot texmax == 1
		double texGenParams[4]={0.0,0.0,0.0,0.0};
		double scl=1.0/((texmax-texmin)[axis]);
		texGenParams[axis]=scl;
		texGenParams[3]=-texmin[axis]*scl;
		glTexGeni(GL_S+axis,GL_TEXTURE_GEN_MODE,GL_OBJECT_LINEAR);
		glTexGendv(GL_S+axis,GL_OBJECT_PLANE,texGenParams);
		glEnable(GL_TEXTURE_GEN_S+axis);
	}
}

/************ Textured OpenGL TerrainRenderer ************/

void TexturedTerrainRenderer::beginTerrain(TerrainCamera &cam_,const TerrainSet *t_) 
{
	cam=&cam_;
	t=t_;
}
void TexturedTerrainRenderer::endTerrain(void) {
	cam=NULL;
	t=0;
	glFlush();
	glDisable(GL_TEXTURE_2D);
}


void TexturedTerrainRenderer::beginTile(const TerrainAddress &addr)
{
	oglCheck("before terrain setup");
	Terrain *tile=t->getTile(addr);
	const osl::GeoImage &geo=tile->getGeo();
	
	if (tile->getTex())
		tile->getTex()->bind(*cam);
	TerrainDrawStart(*cam);
	
	if (0) { // Make image transparent
		glEnable(GL_BLEND);
		glColor4f(1,1,1,0.7);
	}
	
	oglEnable(!(cam->getDrawMode()&drawMode_backface), GL_CULL_FACE);
	glCullFace((geo.pixelSize.x*geo.pixelSize.y)<0?GL_FRONT:GL_BACK);
	
	if (cam->getDrawMode()&drawMode_zbuf) {
		glEnable(GL_DEPTH_TEST);
	} else /* z buffer disabled */ {
		glDisable(GL_DEPTH_TEST);
	}
	glFlush(); /* needed on ATI M9000 Linux, else first triangle's edges flash (!) */
	
	oglCheck("after terrain setup");
}

void TexturedTerrainRenderer::endTile(const TerrainAddress &addr)
{
	flush();
	TerrainDrawStop();
	oglCheck("after terrain cleanup");
}

void TexturedTerrainRenderer::drawObjects(const TerrainLevel *b,const TerrainAddress &addr,TerrainObjectListID objectList)
{
	flush();
	TerrainDrawPause(*cam);
	cam->objectMgr->drawList(cam,b,addr,objectList);
	TerrainDrawResume(*cam);
}

