/**
This is my own take on Peter Lindstrom and Valerio Pascucci's 
dynamic terrain simplification algorithms.

Orion Sky Lawlor, olawlor@acm.org, 2003/12/6, 2006/11/26
*/
#ifndef __OSL_TERRAIN_SIMPLIFY_H
#define __OSL_TERRAIN_SIMPLIFY_H

#include "osl/geo.h"
#include "osl/io.h"
#include "osl/units.h"
#include "osl/raster.h"
#include "osl/geo_tile.h"
#include "terrain/mmap.h"
#include "terrain/utility.h"
#include "terrain/renderer.h"

#include "ogl/util.h"
#include "osl/vector2d.h"
#include "osl/geo_tile.h"

using osl::Vector2d;
using osl::GeoHeight;
typedef osl::Point Point2d;

/**
 Bounding volume for all the geometry built from a vertex.
 
 TerrainBounds don't even know their own 3D location
 (to save storage space), so this value is passed in
 from the caller as "V".
*/
class TerrainBound {
	/// Maximum extent of our (and our children's) geometry from our center.
	float r;
	
	/// Elevation of this point
	float z;
	
	typedef unsigned short error_t;
	/// Maximum world-space error in our volume, in 1.0/eConv meter units.
	///  This is used for distance-based LOD.
	error_t error;
	
	/// Return an error_t value to represent this much distance error, in meters.
	error_t float2error(float e_) {
		enum {eMax=65535}; // integral limit for error_t's
		int ie=(unsigned short)(e_*errorConv+0.99999);
		if (ie>eMax) ie=eMax;
		return (error_t)ie;
	}
	/// If nonzero, we have an interesting child who must be displayed.
	///   If zero, use error opening criterion normally.
	unsigned short mustOpen;
	
public:
	
	/// This is the conversion factor from meters to integral error "e" units;
	///  that is, a k meter height error will be stored as (error_t)(k*errorConv).
	///  Larger values give finer quantization, but less range (e is 0 to 65535).
	enum {errorConv=100};
	
	TerrainBound() {r=0;z=0;error=0;mustOpen=0;}
	
	void setRadius(float r_) {r=r_;}
	void setHeight(float z_) {z=z_;}
	void setError(float e_) {
		error_t ne=float2error(e_);
		if (ne>error) error=ne;
	}
	/// Always open this bound.  Must eventually call decrementOpen.
	void incrementOpen(void) {mustOpen++;}
	void decrementOpen(void) {mustOpen--;}
	
	inline float getRadius(void) const {return r;}
	inline float getHeight(void) const {return z;}
	
	/**
	 Return true if this vertex is "active"-- if the 
	 vertex's subtree should be drawn.  This may include:
	   - Distance and orientation LOD tests. 
	   - View clipping.
	   - Importance measurement.
	 This routine is absolutely speed-critical.
	 
	  \param V The location of this vertex in 3D space.
	  \param c The camera viewing this vertex.
	*/
	bool isActive(const Vector3d &V,const TerrainCamera &c) const {
		if (mustOpen!=0) return true;
		/**
		 Distance LOD test: don't bother if
		    angError <= e / (dist-r)
		    (dist-r) <= e / angError
		    dist <= e / angError+r
		    dist^2 <= (e / angError+r)^2
		*/
		double camSqr=c.getLocation().distSqr(V);
		double tR=r+error*c.getInvError();
		return camSqr <= tR*tR;
	}
	
	/**
	 Add this vertex's bounding volume to our own.
	 Specifically, this establishes the invariant that
	   child.isActive(childV) => this.isActive(V)
	 by modifying our bounding volume.
	 
	 \param V The location of this vertex in 3D space.
	 \param child The child's data structure, which isn't changed.
	 \param childV The location of the child vertex in 3D space.
	*/
	void merge(const Vector3d &V,
		const TerrainBound &child,const Vector3d &childV);
};

/**
 This is the data we keep at each quad.
*/
class TerrainQuad {
public:
	/// Flags controls how this quad is split into triangles.
	typedef unsigned short flags_t;
	flags_t flags;
	
	/// List of objects for my quad.
	TerrainObjectListID objectList;
	
	TerrainQuad() {flags=0; objectList=0; }
};

/**
 * This is a simple 2D table.  The operations are mostly row-centric.
 */
template <class T>
class BasicTable2d {
protected:
	int rows; //Number of entries in table (y height)
	int cols; //Size of each entry in table (x width)
	T *table; //Data in table [rows * cols]
public:
	BasicTable2d(T *src,int cols_,int rows_) 
		:rows(rows_), cols(cols_), table(src) {}
	
	/// "size" of the table is the number of rows:
	inline int size(void) const {return rows;}
	/// Width of the table is the number of columns:
	inline int width(void) const {return cols;}
	
	T *getData(void) {return table;}
	const T *getData(void) const {return table;}
	void setData(T *t) { table = t; }
	
/// Element-by-element operations:
	T &write(int x,int y) {return table[x+y*cols];}
	const T &operator()(int x,int y) const {return table[x+y*cols];}
	
/// Row-by-row operations
	//Get a pointer to a row of the table:
	inline T *getRow(int r) {return &table[r*cols];}
	inline const T *getRow(int r) const {return &table[r*cols];}
	inline T *operator[](int r) {return getRow(r);}
	inline const T *operator[](int r) const {return getRow(r);}
	inline void setRow(int r,T value) {
		T *dest=getRow(r);
		for (int c=0;c<cols;c++) dest[c]=value;
	}
	void set(T value) {
		for (int r=0;r<rows;r++) setRow(r,value);
	}
};

class TerrainLinHeightFn;

/**
  Represents one 2d array of terrain bounding volumes.
  This is the central class of the whole enterprise:
  it's the thing that builds and traverses the bounding volumes.
*/
class TerrainLevel {
	/// Size of our bounding volume table, in vertices.
	int sz;
	/// stride of our table-- multiplier to finest level
	int stride;
	
	/// Converts map coordinates to quads.
	osl::GeoImage geo;
	
	// Our tile number and level
	TerrainAddress myAddr;
	
	/// Our vertex bounding volumes.
	BasicTable2d<TerrainBound> bound;
	/// Dynamically allocated storage for bounding table (0 if from file)
	TerrainBound *alloc_bound;
	
	/// Our quad information. quad(x,y) is bound(x,y) - bound(x+1,y+1).
	BasicTable2d<TerrainQuad> quad;
	/// Dynamically allocated storage for quad information (0 if from file)
	TerrainQuad *alloc_quad;
	
	/// Terrain at one finer/coarser level (0 if we're the finest/coarsest)
	const TerrainLevel *finer, *coarser;
	
	/// Return the height of this (local) vertex
	float getZ(int x,int y) const 
		 {return bound(x,y).getHeight();}
		// {return ht->at(x*stride,y*stride);}
	
	/// Return the 3D location of this (local) coordinate.
	Vector3d getV(int x,int y) const {
		return Vector3d(x*geo.pixelSize.x+geo.origin.x,y*geo.pixelSize.y+geo.origin.y,getZ(x,y));
	}
	
// Error calculation:
	bool clip(Point2d &p) {
		return (p.x<0) || (p.y<0) || (p.x>=sz) || (p.y>=sz);
	}
	
	/// Return the total error, computed using scan conversion, from this triangle.
	///  The points are given at the *finest* resolution.
	void triErrorScan(double &maxErr,
		const TerrainLinHeightFn &h,const GeoHeight *ht,
		const Point2d &a,const Point2d &b,const Point2d &c);
	
	/// Return the total error, computed using a loop, from approximating this terrain with a single triangle.
	void triErrorLoop(double &maxErr,
		const TerrainLinHeightFn &h,const GeoHeight *ht,
		const Point2d &a,const Point2d &b,const Point2d &c);
	
	/// Return the total error from using this triangle.
	void triError(double &maxErr,const GeoHeight *ht,Point2d a,Point2d b,Point2d c);
	
	/// Return the total error from using this quad:
	///   if true, split the a-d axis as triangle(abd)/triangle(dca);
	///   if false, split the b-c axis as triangle(abc)/triangle(bdc).
	double quadError(const GeoHeight *ht,
		const Point2d &a,const Point2d &b,
		const Point2d &c,const Point2d &d,
		bool split);
	
	/// Merge vertex (x,y)'s bounding volume with that of src's (sx,sy)
	void merge(int x,int y, const TerrainLevel &src,int sx,int sy) {
		bound.write(x,y).merge(getV(x,y), src.bound(sx,sy),src.getV(sx,sy));
	}
	/**
	  Compute the approximation errors at *this* level of the 
	  hierarchy-- the height difference that would be present
	  if each center or side vertex isn't activated.
	*/
	void boundHt(const osl::GeoHeight *ht);
	
	/**
	  Propagate the approximation errors from the center vertices
	  of the finer level of the approximation up to our pixels.
	*/
	void boundFinerCenter(void);
	
	/**
	  Propagate the approximation errors from our side vertices
	  to our center vertices.
	*/
	void boundSide(void);
	
// Recursive traversal & rendering:
	/// Return true if this vertex is active.
	inline bool isActive(const TerrainCamera &cam,int x,int y) const {
		return bound(x,y).isActive(getV(x,y),cam);
	}
	
public:
	TerrainLevel(int my_sz,int full_sz,TerrainAddress myAddr_);
	~TerrainLevel();
	
	const BasicTable2d<TerrainBound> &getData(void) const {return bound;}
	inline const TerrainAddress &getAddress(void) const {return myAddr;}
	
	/// Fill our values with this source, starting at source(sx,sy).
	void copy(int sx,int sy,const TerrainLevel &src);
	
	/// Set up this bounding volume.
	void setup(const osl::GeoImage *ht_,
		const TerrainLevel *finer_,const TerrainLevel *coarser_);
	
	/// Allocate space for our bounds (only needed if propagating)
	void allocate(void);
	
	/// Propagate bounds up the bound tree
	void propagate(const osl::GeoHeight *ht);
	
	/// Add this object to our terrain
	void add(TerrainObjectManager *mgr,TerrainObject *obj,const osl::Bbox2d &box);
	
	/// Write this portion of our bounding volume to this file.
	void writeBounds(osl::io::OutputStream &dest,int xs,int xe,int ys,int ye);
	/// Read our bounding volume from this pointer, advancing as appropriate
	const char *readBounds(const char *data);

	inline const osl::GeoImage &getGeo(void) const {return geo;}

	/**
	  Recursively expand the quad (x,y)-(x+2,y+2) to this camera.
	*/
	void expand(TerrainCamera &cam,bool viewCull,int x,int y) const;
	
/* These drawing utilities are useful to renderers */
	
	/**
	  Upload the vertex (x,y) to OpenGL.
	   Output coordinate system is 3D map coordinates.
	*/
	inline void glVertex(int x,int y) const {
		// glVertex3f(x*stride,y*stride,getZ(x,y)); // <- old way
		glVertex3f(
			x*geo.pixelSize.x+geo.origin.x,
			y*geo.pixelSize.y+geo.origin.y,
			getZ(x,y));
		
	}
	
	/// Return the quad (x,y)
	inline const TerrainQuad &getQuad(int x,int y) const {
		return quad(x,y);
	}
	
	/// Return the vertex at x,y
	inline TerrainBound &getBound(int x,int y) {
		return bound.write(x,y);
	}
	
	/**
	This algorithm is based on refining quads, whose vertices 
	fall into three classes:
		corner- pre-existing vertices (vertices)
		center- new quad center vertices (faces)
		side- border vertices (edges)
	<pre>
	  corner  vside   corner
	  hside   center  hside
	  corner  vside   corner
	</pre>
	*/
	typedef enum { 
		vertex_corner=0,
		vertex_vside=1,
		vertex_hside=2,
		vertex_center=3 
	} vertex_kind_t;
	inline vertex_kind_t vertex_kind(int x,int y) const {
		return (vertex_kind_t)((x&1)+((y&1)*2));
	}
	
	/// Return our "stride": our quad multiplier to the finest resolution
	inline int getStride(void) const {return stride;}
	

/* These drawing routines just call cam.renderer to do the real work */

	/// Draw a triangle around vertex (x,y), to vertices v1 and v2.
	void drawTriangle(TerrainCamera &cam,int x,int y,int v1,int v2) const;
	
	/// Draw our quad (x,y), including objects
	void drawQuad(TerrainCamera &cam,int x,int y) const;
	
	/// Draw the objects decorating the quad (x,y)
	inline void drawQuadObjects(TerrainCamera &cam,int x,int y) const {
		drawQuadObjects(cam,quad(x,y),x,y);
	}
	
	/// Draw the objects decorating this quad.
	inline void drawQuadObjects(TerrainCamera &cam,const TerrainQuad &q,int x,int y) const {
		if (q.objectList!=0)
			drawQuadObjectsOutOfLine(cam,q,x,y);
	}
	
	void drawQuadObjectsOutOfLine(TerrainCamera &cam,const TerrainQuad &q,int x,int y) const;
	
};

class TerrainObject;
class TerrainObjectTable;
class TerrainListTable;
class oglTexture;

/// Abstract class: represents everything that can be drawn. (not clear it's useful...)
class TerrainDrawable {
public:
	virtual ~TerrainDrawable();
	virtual void draw(TerrainCamera &cam) =0;
};

/**
 A class for simplifying a height map for rendering.
 This class represents one square tile, which is internally
 represented by a nesting of heirarchical bounding volumes (b).
*/
class Terrain : public TerrainDrawable {
	/** Raw, level 0 height map (used for camera motion, etc.) */
	const osl::GeoImage *ht;
	
	/** Source file for height data */
	osl::io::MmapFile *heightSource;
	
	/** Texture data for our height map.
	  Uses its own texmin, texmax world coordinates, but 
	  these *MUST* cover all our terrain!
	  May be overridden in the TerrainRenderer.
	*/
	TerrainTexture *tex;
	
	// Our tile number
	TerrainAddress myAddr;
	
	/** 
	   Heirarchical bounding volumes.
	     b[0], level 0, has 2x2 vertices, 1x1 quads (used only for the quad)
	     b[1], level 1, has 3x3 vertices, 2x2 quads
	     b[i], level i, has level2size(i) x vertices, (1<<i) x quads
	     b[nLevels] is ht->width x ht->height vertices (full resolution)
	*/
	TerrainLevel *b[33];
	int nLevels;
	void setupLevels(bool propagate);
	
	/// Given a level, return the size in vertices across
	int level2size(int l) { return 1+(1<<l); }
	/// Given the size in vertices (e.g., 33 x 33), return the highest used level.
	int size2level(int sz) {
		int l=1;
		while (level2size(l)<sz) l++;
		if (level2size(l)!=sz) osl::bad("Logic error in Terrain::size2level");
		return l;
	}
public:
	/// Build a completely new Terrain from this height map.
	Terrain(const osl::GeoHeight *ht_);
	
	/// Copy out a window of this existing Terrain.
	///   The window is w x h samples starting at (sx,sy)
	Terrain(const Terrain &src,int w,int h,int sx,int sy,TerrainAddress myAddr_);
	
	/// Read an old Terrain from this directory.
	///  The directory should contain files called "dem.dem" and "out.png"
	Terrain(const osl::String &dirName,TerrainAddress myAddr_);
	
	~Terrain();
	
	/// Add this object to our terrain
	void add(TerrainObjectManager *mgr,TerrainObject *obj,const osl::Bbox2d &box);
	
	/// Return the bound object for this address
	inline TerrainLevel *getBound(const TerrainAddress &addr) const {
		return b[addr.level];
	}
	
	const osl::GeoImage &getGeo(void) const {return *ht;}
	double getHeight(const osl::Vector2d &pixel) const;
	int getMaxLevel(void) const {return nLevels;} ///< *Last* level
	
	/// Set our texture map to this.
	void setTex(TerrainTexture *tex_) {
		delete tex;  tex=tex_;
	}
	/// Get our texture, or NULL if none.
	const TerrainTexture *getTex(void) const {return tex;}
	
	/// Write our bounding volumes to this directory. 
	void write(const osl::String &dirName);
	
	void draw(TerrainCamera &cam,const TerrainAddress &addr);
	void draw(TerrainCamera &cam) {
		draw(cam,TerrainAddress(myAddr.tile));
	}
};

/*********** Texture Input ***************/
/// Read a texture from this file, which includes an extension.
oglTexture *TerrainTexRead(const char *fName);

/// Read from this base name.
oglTexture *TerrainTexReadBase(const char *baseName);

/// Read a RgbaRaster from this base name.
osl::graphics2d::RgbaRaster *TerrainImageReadBase(const char *baseName);


/**
 A tiled set of Terrain objects.
 
 This class exists mostly to work around the texture size limits 
 of graphics cards, but also allows prefetching and memory control.
*/
class TerrainSet : public TerrainDrawable {
	/// Raster pattern of Terrains.  Some Terrains may be NULL,
	///  which indicates their space is missing.
	std::vector<Terrain *> t;
	
	/// Table of objects on the terrain.
	TerrainObjectTable *ot;
	
	/// Bounding box of terrain (map coordinates).
	osl::Bbox2d box;
	
	/// Draw a ring of tiles centered on cen
	void drawTileRing(TerrainCamera &cam,const osl::Point &cen,int radius);
	
	/// Draw tile t (if it exists)
	void drawTile(TerrainCamera &cam,const osl::Point &ti)
	{
		if (ti.x<0 || ti.x>=toTile->width
		  ||ti.y<0 || ti.y>=toTile->height) return; //Out of bounds
		int tile=ti.y*toTile->width+ti.x;
		if (!t[tile]) return;
		draw(cam,tile);
	}
public:
	/// TileSet from map coordinates to terrain tiles.
	osl::TileSet *toTile;
	/// Return the total number of image tiles
	inline int totalTiles(void) const {return toTile->width*toTile->height;}
	
	/// True if we had to partition information ourselves.
	bool needsWrite;
	
	/// Get the terrain image at this tile point.
	Terrain &at(const Point &p) {
		return *t[tileNumberFromPoint(p)];
	}
	
	/// Get tile number (addr.tile) from tile point
	inline int tileNumberFromPoint(const Point &p) {
		return p.y*toTile->width+p.x;
	}
	
	/// Get tile point from tile number (like addr.tile)
	Point tilePointFromNumber(int tileNo) const {
		return Point(tileNo%toTile->width,tileNo/toTile->width);
	}
	
	/**
	 Create a partitioned set of terrains based on this file.
	 Partitions data in memory, and sets "needsWrite", if disk data not partitioned.
	 Reads pre-partitioned partitioned data if it exists on disk. (call "write".)
	*/
	TerrainSet(const char *baseName);
	~TerrainSet();
	
	void draw(TerrainCamera &cam);
	inline void draw(TerrainCamera &cam,const TerrainAddress &addr) {
		t[addr.tile]->draw(cam,addr);
	}
	inline Terrain *getTile(const TerrainAddress &addr) const
		{return t[addr.tile];}
	/// Return the bound object for this address
	inline TerrainLevel *getBound(const TerrainAddress &addr) {
		return t[addr.tile]->getBound(addr);
	}
	/// Return the 2D map projection bounding box
	inline const osl::Bbox2d &getBox2d(void) const {return box;}
	
	/// Partition this height image, which we keep.
	void partition(const osl::GeoHeight &ht);
	
	/// Partition this texture map into this many pieces, 
	////   which are saved as well as written out.
	void partition(osl::graphics2d::RgbaRaster &rast);
	
	/// Add this object to our partitioned terrain.
	///   Our manager then owns the object.
	void add(TerrainObjectManager *mgr,TerrainObject *obj,const osl::Bbox2d &box);
	
	/// Write out our partitioned terrain.
	void write(void);
	
	// These sure don't belong here, but they're used by TerrainController
	osl::Vector2d getOrigin(void);
	osl::Vector2d getCenter(void);
	double getHeight(const osl::Vector2d &atMap,double altitude);
};

// for debugging
void breakhere(void);


#endif
