/*
This code walks the terrain data structure, for example for rendering.

This is my own take on Peter Lindstrom and Valerio Pascucci's 
simplification algorithms.

Orion Sky Lawlor, olawlor@acm.org, 2003/12/6
*/
#include "terrain/terrain.h"
using namespace osl;

/********** Terrain Rendering ****************/

void TerrainSet::draw(TerrainCamera &cam) {
	if (!cam.renderer) osl::bad("Forgot to set renderer in terrain camera!");
	cam.renderer->beginTerrain(cam,this);
	// Draw the terrain tiles in a series of concentric rings 
	//  closing in around the camera.  This seems to be the easiest
	//  way to draw in z-order.
	Point cenTile=toTile->clipPixel(toTile->pixelFmMap(cam.getMap()));
	
	for (int radius=2*(toTile->width-1);radius>=0; radius--)
		drawTileRing(cam,cenTile,radius);
	cam.renderer->endTerrain();
}

/// Draw a ring of tiles centered on cen
void TerrainSet::drawTileRing(TerrainCamera &cam,const osl::Point &cen,int radius)
{
	if (radius==0) drawTile(cam,cen); // A 0-ring is just the center tile
	else { // Draw a line of tiles along each diagonal.
		int i;
		for (i=0;i<radius;i++) drawTile(cam,Point(cen.x-radius+i,cen.y-i));
		for (i=0;i<radius;i++) drawTile(cam,Point(cen.x+i,cen.y-radius+i));
		for (i=0;i<radius;i++) drawTile(cam,Point(cen.x+radius-i,cen.y+i));
		for (i=0;i<radius;i++) drawTile(cam,Point(cen.x-i,cen.y+radius-i));
	}
}

void Terrain::draw(TerrainCamera &cam,const TerrainAddress &addr)
{
	cam.loc2d=ht->pixelFmMap(cam.getMap());
	
	cam.renderer->beginTile(myAddr);
	
// Recursively expand the terrain.
//   The +1 and stuff is because of the way we always draw 4 quads at a time.
	if (addr.level<nLevels)
	  cam.renderer->expand(b[addr.level+1],!(cam.getDrawMode()&drawMode_noclip),
		addr.x*2,addr.y*2);
	else // finest level: just draw the quad
	  b[addr.level]->drawQuad(cam,addr.x,addr.y);

	cam.renderer->endTile(myAddr);
}

/** 
  A "split" is a triangle or recursions; quads are drawn
  as a collection of splits that cover the quad's terrain.

  For the quad:
      (dx,dy)                vertex number
      (0,0)  (1,0)  (2,0)    0  1  2
      (0,1)  (1,1)  (2,1)    7  8  3
      (0,2)  (1,2)  (2,2)    6  5  4
*/
typedef enum {
	split_stop=0,
	// Recursion records: recurse down this diagonal
	recurse_0,recurse_2=recurse_0+2,recurse_4=recurse_0+4,recurse_6=recurse_0+6,
	// Big triangles: from corner to corner
	t_02,t_24=t_02+2,t_46=t_02+4,t_60=t_02+6,
	// Little triangles: from corner to midpoint
	t_01,t_12,t_23,t_34,t_45,t_56,t_67,t_70,
	// Object recursion records: draw objects for this diagonal
	object_0,object_2=object_0+2,object_4=object_0+4,object_6=object_0+6
} split_t;

/**
 The splitTable determines how to divide up a quad
 given the current viewpoint and the current set of active sides.
*/
class splitTable {
public:
	enum { nViews=4 }; // Number of distinct viewpoints (for ordering)
	enum { nSides=16 }; // Number of possible side-active markers
	enum { maxSplits=11 }; // Maximum components for each viewpoint
	
	/**
	 Vertex numbering around a quad: clockwise from topleft.
	*/
	typedef int vertex_t;
	
	splitTable();
	
	/**
	 Return the list of splits for the quad centered at pixel (xCen,yCen)
	 viewed from this camera with active sides as specified.
	*/
	inline const split_t *getSplits(TerrainCamera &cam,int xCen,int yCen, int activeSides) const
	{
		// Decide drawing order based on angle in plane
		//  from camera to center of subdivision (x+1,y+1)
		int viewIdx=
			((cam.loc2d.x<xCen)?0:1)+
			((cam.loc2d.y<yCen)?0:2);
		if (cam.getDrawMode()&drawMode_vieworder) {
			const static Vector3d viewColors[4]={
				Vector3d(1,0,0), Vector3d(0,1,0), Vector3d(0,0,1), Vector3d(0.6,0,0.6)
			};
			glColor3dv(viewColors[viewIdx]);
		}
		return splits[viewIdx][activeSides].splits;
	}
	
	static void getSplits(vertex_t nearVertex,int activeSides,split_t *dest);
private:
	/// One way to split up a quad: terminated by a 0.
	struct quadSplit_t {
		split_t splits[maxSplits];
	};
	quadSplit_t splits[nViews][nSides];
};

splitTable::splitTable() {
	// Map from view number to nearest-to-camera corner vertex:
	const static vertex_t view2vertex[4]={0,2,6,4};
	for (int viewIdx=0;viewIdx<nViews;viewIdx++)
	for (int activeSides=0;activeSides<nSides;activeSides++)
		getSplits(view2vertex[viewIdx],activeSides,splits[viewIdx][activeSides].splits);
}
void splitTable::getSplits(int nearVertex,int activeSides,split_t *splits)
{
	enum {nVertex=8};
	// Mark the active vertices:
	bool activeVertex[nVertex];
	for (int i=0;i<nVertex;i++) 
		activeVertex[i]=((i&1)==0); // Mark the corners as active
	if (activeSides&1) activeVertex[7]=true;
	if (activeSides&2) activeVertex[1]=true;
	if (activeSides&4) activeVertex[3]=true;
	if (activeSides&8) activeVertex[5]=true;
	
	/**
	  We want to generate the split records ordered so that overdraw
	  will perform our z-sort.  The way we do this is based on corner
	  vertices-- we loop over the 4 corners from back to front,
	  and output the fragments that correspond to each corner.
	*/
	for (int vi=0;vi<4;vi++) {
		/* source vertex */
		int v=-999;
		bool useLong=false;
		switch(vi) {
		case 0: v=(nearVertex+nVertex/2)%nVertex; useLong=true; break; // far
		case 1: v=(nearVertex+nVertex/4+nVertex)%nVertex; break; // left
		case 2: v=(nearVertex-nVertex/4+nVertex)%nVertex; break; // right
		case 3: v=nearVertex%nVertex; useLong=true; break; // near
		};
		int pv=(v-1+nVertex)%nVertex, nv=(v+1+nVertex)%nVertex;
		if (activeVertex[pv] && activeVertex[nv]) 
		{ // Both edges around corner are active-- recurse
			*splits++=(split_t)(recurse_0+v);
		}
		else 
		{ // Not all edges are active-- split corner into triangles:
			// preceeding span: long or short triangles
			if (activeVertex[pv]) 
				*splits++=(split_t)(t_01+pv); // active-- use short triangle
			else if (useLong) {
				int otherVertex=(v-2+nVertex)%nVertex;
				*splits++=(split_t)(t_02+otherVertex);
				if (vi==3) *splits++=(split_t)(object_0+otherVertex);
			}
			
			// next span: long or short triangles
			if (activeVertex[nv]) 
				*splits++=(split_t)(t_01+v); // active-- use short triangle
			else if (useLong) {
				int otherVertex=(v+2+nVertex)%nVertex;
				*splits++=(split_t)(t_02+v);
				if (vi==3) *splits++=(split_t)(object_0+otherVertex);
			}
			
			// Delay object records on top of long triangles until closest vertex
			if (vi==1 && !activeVertex[pv]) continue;
			if (vi==2 && !activeVertex[nv]) continue;
			
			// Object record for this vertex
			*splits++=(split_t)(object_0+v);
		}
	}
	*splits++=split_stop;
}

/**
  Draw a vertex fan triangle between the center (x+1,y+1), v1, and v2.
  v1 and v2 are a local "vertex number", organized clockwise around
  the center (x+1,y+1).
*/
inline void TerrainLevel::drawTriangle(TerrainCamera &cam,int x,int y,int v1,int v2) const
{
	TerrainAddress addr=myAddr;
	addr.x=x; addr.y=y;
	cam.renderer->drawTriangle(this,addr,v1,v2);
}

/**
  Draw the quad (x,y), including objects.
*/
inline void TerrainLevel::drawQuad(TerrainCamera &cam,int x,int y) const
{
	TerrainAddress addr=myAddr;
	addr.x=x; addr.y=y;
	cam.renderer->drawQuad(this,addr);
}

/// Draw the objects decorating the quad (x,y)
inline void TerrainLevel::drawQuadObjectsOutOfLine(TerrainCamera &cam,const TerrainQuad &q,int x,int y) const
{
	TerrainAddress addr=myAddr;
	addr.x=x; addr.y=y;
	cam.renderer->drawObjects(this,addr,q.objectList);
}

/**
  Render the quad (x,y)-(x+2,y+2) to this camera.
*/
void TerrainLevel::expand(TerrainCamera &cam,bool viewCull,int x,int y) const {
	const TerrainBound &c=bound(x+1,y+1);
	Vector3d cV=getV(x+1,y+1);
	
// View cull center vertex:
	if (viewCull) {
		cam.view(TerrainCamera::viewCullTrue);
		osl::boundview_t view=cam.sphereInView(cV,c.getRadius());
		if (view==osl::boundview_none) 
		{ // We're totally out of bounds
			// Some of our objects might still be in bounds:
			coarser->drawQuadObjects(cam,x/2,y/2);
			return; 
		}
		if (view==osl::boundview_all)
			viewCull=false; // Don't bother checking bounds any more
	}
	else
		cam.view(TerrainCamera::viewCullFalse);
	
// Active tests:
	if (!c.isActive(cV,cam)) 
	{ /* we are inactive--draw coarser quad and leave */
		cam.view(TerrainCamera::viewInactive);
		coarser->drawQuad(cam,x/2,y/2);
		return;
	}
	
	/*
	Test each side vertex, and build a bit mask of active sides:
	      2
	   1     4
	      8
	*/
	int activeSides=
		(isActive(cam,x  ,y+1)?1:0)+
		(isActive(cam,x+1,y  )?2:0)+
		(isActive(cam,x+2,y+1)?4:0)+
		(isActive(cam,x+1,y+2)?8:0);
	// Collect view statistics:
	if (finer) cam.view(activeSides);
	else cam.view(TerrainCamera::viewSides+activeSides);
	
// Split up this quad and draw fragments:
	const static splitTable splits;
	const split_t *s=splits.getSplits(cam,(x+1)*stride,(y+1)*stride, activeSides);

	if (!(cam.getDrawMode()&drawMode_crack)) { /* non-crack version */
		for (int i=0;s[i];i++) 
		switch(s[i]) {
		case split_stop: goto end_splitloop;

		case recurse_0:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+0),2*(y+0));
			else drawQuad(cam,x+0,y+0);
			break;
		case recurse_2:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+1),2*(y+0));
			else drawQuad(cam,x+1,y+0);
			break;
		case recurse_4:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+1),2*(y+1));
			else drawQuad(cam,x+1,y+1);
			break;
		case recurse_6:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+0),2*(y+1));
			else drawQuad(cam,x+0,y+1);
			break;

		case object_0: drawQuadObjects(cam,x+0,y+0); break;
		case object_2: drawQuadObjects(cam,x+1,y+0); break;
		case object_4: drawQuadObjects(cam,x+1,y+1); break;
		case object_6: drawQuadObjects(cam,x+0,y+1); break;

		case t_02: drawTriangle(cam,x,y, 0,2); break;
		case t_24: drawTriangle(cam,x,y, 2,4); break;
		case t_46: drawTriangle(cam,x,y, 4,6); break;
		case t_60: drawTriangle(cam,x,y, 6,0); break;

		case t_01: drawTriangle(cam,x,y, 0,1); break;
		case t_12: drawTriangle(cam,x,y, 1,2); break;
		case t_23: drawTriangle(cam,x,y, 2,3); break;
		case t_34: drawTriangle(cam,x,y, 3,4); break;
		case t_45: drawTriangle(cam,x,y, 4,5); break;
		case t_56: drawTriangle(cam,x,y, 5,6); break;
		case t_67: drawTriangle(cam,x,y, 6,7); break;
		case t_70: drawTriangle(cam,x,y, 7,0); break;
		};
		end_splitloop: ;
	}
	else 
	{ /* Version showing cracks caused by all-quad rendering */
		// printf("Now on crack...\n");
		for (int i=0;s[i];i++) 
		switch(s[i]) {
		case split_stop: goto end_splitloop2;

		case recurse_0:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+0),2*(y+0));
			else drawQuad(cam,x+0,y+0);
			break;
		case recurse_2:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+1),2*(y+0));
			else drawQuad(cam,x+1,y+0);
			break;
		case recurse_4:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+1),2*(y+1));
			else drawQuad(cam,x+1,y+1);
			break;
		case recurse_6:
			if (finer) cam.renderer->expand(finer,viewCull,2*(x+0),2*(y+1));
			else drawQuad(cam,x+0,y+1);
			break;

		case object_0: drawQuad(cam,x+0,y+0); break;
		case object_2: drawQuad(cam,x+1,y+0); break;
		case object_4: drawQuad(cam,x+1,y+1); break;
		case object_6: drawQuad(cam,x+0,y+1); break;
		default: break;
		};
		end_splitloop2: ;
	}
}

