/**
Utility code for terrain rendering.

Orion Sky Lawlor, olawlor@acm.org, 2006/11/26
*/
#include "terrain/terrain.h"
using namespace osl;

/********* TerrainSet **************/

osl::Vector2d TerrainSet::getOrigin(void) {
	const GeoImage &g=t[0]->getGeo();
	return g.mapFmPixel(osl::Vector2d(0,0));
}
osl::Vector2d TerrainSet::getCenter(void) {
	Vector2d sum(0,0); double n=0;
	for (unsigned int i=0;i<t.size();i++) 
	if (t[i]) {
		const GeoImage &g=t[i]->getGeo();
		sum+=g.mapFmPixel(osl::Vector2d(g.width/2,g.height/2));
		n+=1.0;
	}
	return sum*(1.0/n);
}
double TerrainSet::getHeight(const osl::Vector2d &map,double altitude) {
	if (altitude>10.0) { // Expand high samples with a larger filter:
		double d=altitude*0.1;
		double h=0;
		return 0.2*(
			getHeight(osl::Vector2d(map.x+d,map.y),h)+
			getHeight(osl::Vector2d(map.x-d,map.y),h)+
			getHeight(osl::Vector2d(map.x,map.y+d),h)+
			getHeight(osl::Vector2d(map.x,map.y-d),h)+
			getHeight(osl::Vector2d(map.x,map.y),h)
		);
	}
	Point t=toTile->pixelFmMap(map);
	if (t.oob(toTile->width,toTile->height)) {
		// Out-of-bounds height:
		return at(Point(0,0)).getHeight(Vector2d(0,0));
	}
	Vector2d pixel=at(t).getGeo().pixelFmMapd(map);
	return at(t).getHeight(pixel);
}
double Terrain::getHeight(const osl::Vector2d &pixel) const {
	// bilinear interpolation:
	int ix=(int)pixel.x, iy=(int)pixel.y;
	double dx=pixel.x-ix, dy=pixel.y-iy;
	double tl=b[nLevels]->getData()(ix  ,iy  ).getHeight();
	double tr=b[nLevels]->getData()(ix+1,iy  ).getHeight();
	double bl=b[nLevels]->getData()(ix  ,iy+1).getHeight();
	double br=b[nLevels]->getData()(ix+1,iy+1).getHeight();
	double t=tl+dx*(tr-tl);
	double b=bl+dx*(br-bl);
	return t+dy*(b-t);
}


/********** TerrainCamera *************/

/// Return the angle, in radians, between pixel (0,0) 
///   and this pixel on this viewpoint.
double angleToOrigin(const osl::Viewpoint &vp,int x,int y) {
	return acos(
	  vp.getPixelRay(osl::Vector2d(0,0)).dir
		.cosAng(
	  vp.getPixelRay(osl::Vector2d(x,y)).dir
		)
	);
}

TerrainCamera::TerrainCamera(const osl::Viewpoint &vp_,
	double pixelErrTol,int drawMode_)
	:drawMode(drawMode_), 
	 renderer(0), objectMgr(0),
	 vp(vp_)
{
	loc2d=Point(0,0);
	int cx=vp.getWidth()/2, cy=vp.getHeight()/2;
	osl::Vector3d viewCenter=vp.viewplane(osl::Vector2d(cx,cy));
	if (vp.isOrthographic())
		loc=Vector3d(viewCenter+vp.getZ().dir()*10000.0);
	else /* perspective camera */
		loc=Vector3d(vp.getEye());
	viewDir=Vector3d(vp.getZ());
	
	terrainTex=NULL;
	
	for (int i=0;i<viewLast;i++) viewFreq[i]=0;
	
	/*
	Angular error tolerance: computed from the 
	  angular size of one pixel, as computed by 
	  comparing the ray directions from (0,0) to (1,0).
	
	FIXME: doesn't work for skew or orthogonal cameras...
	*/
	double errTol=pixelErrTol*acos(
		(viewCenter-osl::Vector3d(loc))
	.cosAng(vp.viewplane(osl::Vector2d(cx+1,cy))-osl::Vector3d(loc))
	);
	invErr=(1.0/errTol)*(1.0/TerrainBound::errorConv);
	
	/// View culling parameters:
	for (int c=0;c<nClip;c++) {
		osl::Halfspace3d h=vp.getClip(c);
		clip[c].n=Vector3d(h.n);
		clip[c].d=h.d;
		clip[c].normalize();
	}
}

TerrainCamera::~TerrainCamera() {
}

void TerrainCamera::viewStats(void)
{
	int i,n=0;
	for (i=0;i<=viewInactive;i++) n+=viewFreq[i];
	const char *fmt="%2.0f%% ";
	printf("Side Freq:       ");
	for (i=0;i<viewSides;i++)
		printf(fmt,100.0*viewFreq[i]/n);
	printf("\nFinal Side Freq: ");
	for (i=0;i<viewSides;i++)
		printf(fmt,100.0*viewFreq[viewSides+i]/n);
	printf("\nInactive: ");
	printf(fmt,100.0*viewFreq[viewInactive]/n);
	
	int t=viewFreq[viewCullTrue], f=viewFreq[viewCullFalse];
	n=t+f;
	int nQuad=viewFreq[viewQuads];
	int nPoly=viewFreq[viewPolys]+2*nQuad;
	int nVertex=viewFreq[viewVertices]+4*nQuad;
	printf("\nView clip tested %.0f%% / %d opened / %d polygons, %.0f%% quads / %.2f vertex/poly\n\n",
		100.0*t/n, n, 
		nPoly, (100.0*(2*nQuad))/nPoly,
		((double)nVertex)/nPoly
		);
}

TerrainDrawable::~TerrainDrawable() {}

/// Print address (for debugging)
void TerrainAddress::print(void) const {
	printf("tile%d level%d (%d,%d)\n",tile,level,x,y);
}

/******** Terrain Objects ********
  Default implementation is to have no terrain objects.
*/
TerrainObjectManager::~TerrainObjectManager() {}

void TerrainObjectManager::addObject(TerrainObject *obj)
{
	printf("ERROR-- base TerrainObjectManager can't handle object additions!\n");
	abort();
}

TerrainObjectListID TerrainObjectManager::appendList(
	const TerrainAddress &addr,
	TerrainObjectListID oldList,TerrainObject *obj)
{
	return 0;
}
	
void TerrainObjectManager::drawList(const TerrainCamera *cam,const TerrainLevel *b,
	const TerrainAddress &addr,TerrainObjectListID objectList) { }
