/*
Terrain reading, partitioning, and writing.

This is my own take on Peter Lindstrom and Valerio Pascucci's 
simplification algorithms.

Orion Sky Lawlor, olawlor@acm.org, 2004/1/31
*/
#include "osl/raster.h"
#include "terrain/terrain.h"
using namespace osl;

/// Return the filename of the i'th piece of this terrain
osl::io::File TerrainPieceName(const char *baseName,int i,const char *ext)
{
	char tmp[1000];
	sprintf(tmp,"%s_partitioned/%d",baseName,1+i);
	return osl::io::File(tmp,ext);
}

/// Compensate this height field for the curvature of the earth,
///  by decreasing the heights as appropriate
void drapeOverEarth(GeoHeight *ht,double r_ea)
{
	return; /* <- over large enough areas, draping distorts normal vector and elevations
		FIXME: update GUI to deal with this effect
			- Vary the up vector
			- Vary the 
	 */

	// Zero point of drape: middle of image
	Vector2d z=ht->mapFmPixel(Vector2d(ht->width/2,ht->height/2));
	for (int y=0;y<ht->height;y++)
	for (int x=0;x<ht->width;x++) {
		/*
		   On circle, x^2+y^2=r^2 so
		      y=sqrt(r^2-x^2)
		   Here, x is distance in map, r radius of earth, y is z-drop.
		*/
		double mapDistDiff=r_ea*r_ea-(ht->mapFmPixel(x,y)-z).magSqr();
		if (mapDistDiff<0) mapDistDiff=0;
		double dz=sqrt(mapDistDiff)-r_ea;
		ht->at(x,y)+=dz;
	}
}


/***************** Partitioning *****************/

/* return this filename without any extensions. */
std::string clipExtension(std::string name) {
	int i=name.size()-1;
	while (i>0 && name[i]!='.' && name[i]!='/' && name[i]!='\\') i--;
	if (i>0 && name[i]=='.') return std::string(name,0,i);
	else return name; /* whole thing */
}

TerrainSet::TerrainSet(const char *baseName) 
	:toTile(0)
{
/* Trace-Only  (for debugging) */
#define TO(x) /* x */
	TO( printf("   terrain: checking for .geo\n"); fflush(stdout); )
	String dirName(clipExtension(baseName));
	
	dirName+="_partitioned";
	
	// Partition if necessary:
	if (!osl::io::File(dirName+".geo").exists()) {
		needsWrite=true;
		
		GeoHeight ht(osl::io::File(baseName,"dem"));
		// Try to cut into reasonable-sized pieces:
		const int terrainSize=128;
		int nDiv=ht.width/terrainSize; // Number of pieces along each axis
		if (nDiv<1) nDiv=1;
		else if (nDiv>10) nDiv=8;
		else if (nDiv>6) nDiv=8;
		else if (nDiv>2) nDiv=4;
		
		GeoImage tileGeo(ht.origin,
			Vector2d(ht.pixelSize.x*(ht.width-1),
			         ht.pixelSize.y*(ht.height-1))*(1.0/nDiv),
			nDiv, nDiv);
		toTile=new osl::TileSet(tileGeo,dirName);
		
		double r_ea=6378000; // Approximate radius of earth, meters
		drapeOverEarth(&ht,r_ea);
		
		/// Partition heights
		partition(ht);
		
		// Must write out tile array before partitioning image, to make parent directories
		toTile->write();
		
		// Partition image
		osl::graphics2d::RgbaRaster *img=TerrainImageReadBase(baseName);
		if (img) {
			partition(*img);
			delete img;
		}
	}
	else { // Directory already exists: read in partitioned data
		needsWrite=false;
		
		TO(printf("   terrain: reading tiles\n"); fflush(stdout); )
		toTile=new osl::TileSet(dirName);
		for (int y=0;y<toTile->height;y++)
		for (int x=0;x<toTile->width;x++) {
			printf("   terrain: reading tile %d,%d\n",x,y); fflush(stdout);
			Point p(x,y);
			Terrain *tile=NULL;
			//if (toTile->exists(p))  //<- this is dumb, and breaks on Windows.
				tile=new Terrain(toTile->directory(p),TerrainAddress(x+y*toTile->width));
			t.push_back(tile);
		}
		TO(printf("   terrain: read all tiles\n"); fflush(stdout); )
	}
	
	TO(printf("   terrain: computing bounding box\n"); fflush(stdout); )
	box.empty();
	for (unsigned int i=0;i<t.size();i++) {
		TO(printf("   terrain: computing bounding box for %d\n",i); fflush(stdout); )
		Terrain *tile=t[i];
		TO(printf("   terrain: loaded tile %p\n",tile); fflush(stdout); )
		const osl::GeoImage &geo=tile->getGeo();
		TO(printf("   terrain: loaded geo %p\n",geo); fflush(stdout); )
		const osl::Bbox2d &b=geo.getBox();
		TO(printf("   terrain: loaded box %p\n",b); fflush(stdout); )
		box=box.getUnion(b);
	}
	TO(printf("   terrain: done with split\n"); fflush(stdout); )
	
}
TerrainSet::~TerrainSet() {
	for (unsigned int i=0;i<t.size();i++)
		delete t[i];
	delete toTile;
}


/// Return the largest power of two less than or equal this value.
int floorPowerTwo(int v) {
	int d=1;
	while (d<=v) d*=2;
	return d/2;
}

/// Partition this height image into this many pieces, which we keep.
void TerrainSet::partition(const osl::GeoHeight &ht) 
{
	printf("Read %d x %d elevation map: partitioning\n",ht.width,ht.height);
	
	/// Build a master terrain image containing everything.
	///  This is a hack: it keeps us from having to deal with 
	///  partitioned error propagation.
	Terrain master(&ht);
	
	/// Cut up the master image into little pieces.
	int wsz=(ht.width-1)/toTile->width;
	for (int py=0;py<toTile->height;py++) 
	for (int px=0;px<toTile->width;px++) 
	{
		Terrain *sub=new Terrain(master,wsz+1,wsz+1,px*wsz,py*wsz,
			TerrainAddress(px+py*toTile->width));
		t.push_back(sub);
	}
}

/// Make a matrix mapping (0,0) to tl, and (img.wid,img.ht) to br:
Matrix2d makeMapFmRaster(const osl::graphics2d::Raster &img,Vector2d tl,Vector2d br)
{	
	Matrix2d ret(1.0);
	ret.setO(tl);
	ret.setX(Vector2d((br.x-tl.x)/img.wid,0));
	ret.setY(Vector2d(0,(br.y-tl.y)/img.ht));
	return ret;
}

/// Partition this texture map into this many pieces, which are saved and written out.
void TerrainSet::partition(osl::graphics2d::RgbaRaster &img)
{
	/**
	 Compute an "effective width"-- a power of two width
	 (because textures must be powers of two) just under the 
	 terrain size (which is always one more than a power of two).
	*/
	int eWid=floorPowerTwo(img.wid);
	printf("Read %d x %d image: partitioning as %d x %d\n",img.wid,img.ht,eWid,eWid);
	int wsz=eWid/toTile->width; /* Size of subdivided image chunks */
	
	/* Find the overall map-coordinates bounding box */
	Vector2d tl=at(Point(0,0)).getGeo().mapFmPixel(0,0);
	const GeoImage &geo=at(Point(toTile->width-1,toTile->height-1)).getGeo();
	Vector2d br=geo.mapFmPixel(geo.width-1,geo.height-1);
	
	/* Build matrices to convert from source image pixels to map coordinates: */
	Matrix2d mapFmPix=makeMapFmRaster(img,tl,br), pixFmMap(1.0);
	
	/* FIXME: read and apply a bitmap geolocation correction matrix to mapFmPix here */
	
	mapFmPix.invert(pixFmMap);
	
	for (int py=0;py<toTile->height;py++) 
	for (int px=0;px<toTile->width;px++) {
		/* Find the map coordinates of this tile */
		const GeoImage &geo=at(Point(px,py)).getGeo();
		Vector2d tl=geo.mapFmPixel(0,0); 
		Vector2d br=geo.mapFmPixel(geo.width-1,geo.height-1);
		
		/* Expand coordinates by half a pixel (for nice bilinear interpolation) */
		double expand=0.5;
		Vector2d d((br.x-tl.x)/(wsz-2*expand),(br.y-tl.y)/(wsz-2*expand)); /* One pixel size, in map coordinates */
		tl-=d*expand; br+=d*expand;
		
		/* Extract a portion of the big image covering tl-br */
		osl::graphics2d::Rasterizer subR(wsz,wsz);
		subR.clear(osl::graphics2d::Color::clear);
		Matrix2d mapFmSub=makeMapFmRaster(subR.getBuffer(),tl,br);
		Matrix2d pixFmSub, subFmPix;
		pixFmMap.product(mapFmSub,pixFmSub);
		pixFmSub.invert(subFmPix); /* Returns subimage pixels given big image pixels */
		osl::graphics2d::GraphicsState gs;
		gs.setMatrix(subFmPix);
		subR.copy(gs,img);
		osl::graphics2d::Raster &sub=subR.getBuffer();
		
		std::string texBase=toTile->createDirectory(Point(px,py))+"out";
		sub.write((texBase+".jpg").c_str());
		GeoImage texGeo(tl,
			(br-tl)/wsz,
			wsz,wsz);
		texGeo.write((texBase+".geo").c_str());
		
		at(Point(px,py)).setTex(new TerrainTexture(
			makeOglTexture(*(osl::graphics2d::RgbaRaster *)&sub),
			texGeo.mapFmPixel(0,0),
			texGeo.mapFmPixel(texGeo.width,texGeo.height)
		));
	}
}

/**
  Copy out an area of the source image starting at (sx,sy) and (w x h)
  into dest, which should be (w x h).
*/
void copyData(int w,int h,int sx,int sy,
	const BasicTable2d<TerrainBound> &src,
	BasicTable2d<TerrainBound> &dest)
{
	for (int y=0;y<h;y++)
	for (int x=0;x<w;x++)
		dest.write(x,y)=src(sx+x,sy+y);
}

/// Copy out a window of this existing Terrain.
///   The window is w x h samples starting at (sx,sy)
Terrain::Terrain(const Terrain &src,int w,int h,int sx,int sy,TerrainAddress myAddr_)
	:heightSource(0), tex(0)
{
	myAddr=myAddr_;
	GeoImage *nht=new GeoImage(*src.ht);
	nht->origin=nht->mapFmPixel(sx,sy);
	nht->width=w; nht->height=h;
	ht=nht;
	setupLevels(false);
	
	/// Copy out data for each corresponding level:
	for (int up=0;up<=nLevels;up++) {
		int div=(1<<up);
		b[nLevels-up]->allocate();
		b[nLevels-up]->copy(sx/div,sy/div, *src.b[src.nLevels-up]);
	}
}

void TerrainLevel::copy(int sx,int sy,const TerrainLevel &src) {
	int x,y;
	for (y=0;y<sz;y++)
	for (x=0;x<sz;x++)
		bound.write(x,y)=src.bound(sx+x,sy+y);
	for (y=0;y<sz-1;y++)
	for (x=0;x<sz-1;x++)
		quad.write(x,y)=src.quad(sx+x,sy+y);
}


/***************** Terrain Output ***************/

/// Write out ".trn" files for our terrain.
void TerrainSet::write(void) {
	for (int py=0;py<toTile->height;py++) 
	for (int px=0;px<toTile->width;px++) {
		at(Point(px,py)).write(toTile->createDirectory(Point(px,py)));
	}
}

/// Divide up our bounding volumes and write to these files.
void Terrain::write(const osl::String &dirName) {
	osl::io::FileOutputStream f(dirName+"dem.trn");
	for (int i=0;i<=nLevels;i++) {
		int wsz=(level2size(i)-1);
		int x=0, y=0; // Formerly: x=wsz*px, y=wsz*py;
		b[i]->writeBounds(f,  x,x+wsz,  y,y+wsz);
		if (i==nLevels) { /* now at finest level-- write .geo file too */
			GeoImage geo(ht->mapFmPixel(Point(x,y)),ht->pixelSize,
				wsz+1,wsz+1);
			geo.write((dirName+"dem.geo").c_str());
		}
	}
}

/// Write this portion of our bounding volume to this file.
void TerrainLevel::writeBounds(osl::io::OutputStream &dest,int xs,int xe,int ys,int ye) {
	// printf("   writing (%d x %d): %d,%d - %d,%d\n", sz,sz, xs,ys, xe,ye);
	int y;
	// Write vertex bounding information
	for (y=ys;y<=ye;y++)
		dest.write(bound[y]+xs,sizeof(TerrainBound)*(xe+1-xs));
	// Write quad bound information
	for (y=ys;y<ye;y++)
		dest.write(quad[y]+xs,sizeof(TerrainQuad)*(xe-xs));
}



/***************** Terrain Input **************/
Terrain::Terrain(const osl::String &dirName,TerrainAddress myAddr_)
	:myAddr(myAddr_)
{
	ht=new GeoImage((dirName+"dem.dem").c_str());
	tex=new TerrainTexture(
		TerrainTexReadBase((dirName+"out").c_str()),
		GeoImage((dirName+"out").c_str())),
	setupLevels(false);
	
	int i, len=0;
	for (i=0;i<=nLevels;i++) {
		int sz=level2size(i);
		len+=sizeof(TerrainBound)*sz*sz;
		len+=sizeof(TerrainQuad)*(sz-1)*(sz-1);
	}
	
	heightSource=new osl::io::MmapFile(
		(dirName+"dem.trn").c_str(), len, osl::io::MmapRead);
	
	char *data=(char *)heightSource->getData();
	const char *src=data;
	for (i=0;i<=nLevels;i++) src=b[i]->readBounds(src);
	if (src!=data+len) osl::bad("Terrain::Terrain: len/src mismatch");
}

/// Read our bounding volume from this pointer, advancing as appropriate
const char *TerrainLevel::readBounds(const char *data)
{
	bound.setData((TerrainBound *)data);
	data+=sizeof(TerrainBound)*sz*sz;
	quad.setData((TerrainQuad *)data);
	data+=sizeof(TerrainQuad)*(sz-1)*(sz-1);
	return data;
}

/*********** Texture Input ***************/

/// Read this image file from this basename, or return NULL if none is readable.
osl::graphics2d::RgbaRaster *TerrainImageReadBase(const char *baseName)
{
	const int nExt=4;
	const char *ext[nExt]={"png","ppm","tga","jpg"};
	osl::graphics2d::RgbaRaster *ret=new osl::graphics2d::RgbaRaster();
	for (int e=0;e<nExt;e++) {
		try {
			ret->read(osl::io::File(baseName,ext[e]).toString());
			return ret;
		} catch (Exception *e) {
			delete e;
		}
	}
	delete ret;
	return NULL;
}

oglTexture *TerrainTexRead(const char *fName) {
	try {
		osl::graphics2d::RgbaRaster rast; 
		rast.read(fName);
		return makeOglTexture(rast);
	} catch (Exception *e) {
		printf("Error reading texture '%s'\n",fName);
		e->printStackTrace();
		abort();
		delete e;
	}
	return NULL;
}

oglTexture *TerrainTexReadBase(const char *baseName) 
{
	osl::graphics2d::RgbaRaster *rast=TerrainImageReadBase(baseName);
	if (rast==NULL) return NULL;
	oglTexture *ret=makeOglTexture(*rast);
	delete rast;
	return ret;
}




