/*
Various funny renderers for terrain code.
Main TerrainRenderer declaration is in terrain/camera.h

Orion Sky Lawlor, olawlor@acm.org, 2004/10/28
*/
#ifndef __OSL_TERRAIN_RENDERER_H
#define __OSL_TERRAIN_RENDERER_H

#include "ogl/util.h" /* for oglTexture */

/*********** Rendering utilities ***************/
/**
  Prepare to draw this texture over the subsequent terrain.
    \param 
  The texture will be stretched so texture coordinate (0,0)
  lies at world coordinate texmin, and texture coordinate 
  (1,1) lies at world coordinate texmax.
*/
class TerrainTexture {
	const oglTexture *tex; ///< Texture to apply
	Vector2d texmin,texmax; ///< World coordinates of texture corners
public:
	TerrainTexture(const oglTexture *tex_,
		const Vector2d &texmin_,const Vector2d &texmax_);
	TerrainTexture(const oglTexture *tex_,
		const osl::GeoImage &texGeo);
	
	/// Begin using our texture on this camera.
	///  Overwrites previous texture.
	void bind(TerrainCamera &cam) const;

	/// Quickly rebind this texture (after a previous real bind)
	inline void bind(void) const { tex->bind(); }
};

/// Temporarily stop drawing terrain, prepare to draw objects.
///  Turns off the texture coordinate generation used for terrain.
void TerrainDrawPause(TerrainCamera &cam);

/// Resume drawing terrain after drawing objects.
///  Turns on the texture coordinate generation used for terrain.
void TerrainDrawResume(TerrainCamera &cam);

/*********** Renderers **************/
/**
  Renders terrain using simple OpenGL vertex calls, in geographic coordinates (m).
  You need to set up the texturing, lighting, etc beforehand.
*/
class VertexTerrainRenderer : public TerrainRenderer {
public:
	VertexTerrainRenderer() {}
	
	virtual void endTile(const TerrainAddress &addr);
	virtual void drawQuad(const TerrainLevel *b,const TerrainAddress &addr);
	virtual void drawTriangle(const TerrainLevel *b,const TerrainAddress &addr,int v1,int v2);

protected:
	/* Eliminates any queued-up triangles (calls glEnd()).
	  Only needed if you want to draw new geometry 
	  before the terrain finishes rendering (before endTile).
	*/
	void flush(void);
};

/**
  Renders terrain using OpenGL textures (as usual)
*/
class TexturedTerrainRenderer : public VertexTerrainRenderer {
protected:
	const TerrainSet *t;
public:
	TexturedTerrainRenderer()
		:t(0) {}
	
	virtual void beginTerrain(TerrainCamera &cam,const TerrainSet *s);
	virtual void endTerrain(void);
	virtual void beginTile(const TerrainAddress &addr);
	virtual void endTile(const TerrainAddress &addr);
	virtual void drawObjects(const TerrainLevel *b,const TerrainAddress &addr,TerrainObjectListID list);
};

#endif
